/* Scanner template, you can either:*/
/* - Copy this file to another file and change it.*/
/* - Change this file directly, it wont be overwritten.*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include "ttterm.h"
#include "ttbuffer.h"
#include "ttutils.h"
#include "ttparse.h"
#include "TIGERScan.h"
#include "TIGERParse.h"

#define MAX_KW_SIZE 1000

static BOOL is_whitespace(int ch)
{
	return ch <= ' ' && ch != '\0';
}

TTterm TTparseTIGER_LAYOUT(void)
{
    const char *buf = TTparse_get_string();
    int len;
    int old_len;
    
    len = 0;
    old_len = len - 1;

    while (len > old_len)
    {
        old_len = len;
        if (is_whitespace(buf[len])) {
            len++;
        } else if (buf[len] == '(' && buf[len+1] == '*') {
            len += 2;
            while (!(buf[len] == '*' && buf[len+1] == ')')) {
                if (buf[len] == '\0') TTparse_error();
                len++;
            }
            len += 2;
        }
    }

    return TTparse_word((size_t)len);
}

/* There are basically two choices here:*/
/* - Keep layout in the parsetree, as is done here using "sandwich".*/
/* - Ignore layout, this will give a more compact parsetree, but*/
/*   preservation of layout will not be possible.*/
static TTterm scan_word(size_t len)
{
    TTterm t1, t2;

    t1 = TTparse_word(len);
    t2 = TTparseTIGER_LAYOUT();
    return TTbuild_sandwich(TTbuild_empty_string(), t1, t2);
}

/* Addapt this code to follow the rules for keywords and identifiers in the. */
/* target language. */
TTterm TTparseTIGER_KEYWORD(const char * name)
{
    const char *buf;
    size_t len = 0;
    BOOL possible_identifier = FALSE;
    int ch, ch2;
    buf = TTparse_get_string();
    while ((ch = name[len]) != '\0') {
    	ch2 = buf[len];
    	if (ch2 == 0) return 0;
    	if (ch != ch2) return 0;
    	if (isalnum(ch)) possible_identifier = TRUE;
    	len++;
    }
    ch = buf[len];
    if (possible_identifier && (isalnum(ch) || ch == '_'))
    	return 0;
    return scan_word(len);
}

/* Passing the calls through to TTparse_disjunction and TTparse_conjunction, */
/* as is done here, should be sufficient for most languages.*/
/* See the COBOL parser for a counter example.*/
TTterm TTparseTIGER_DISJUNCTION(TTfun fun, size_t n, TTterm(**parsers) (void))
{
    TTterm result = TTparse_disjunction(fun, n, parsers);
/*
fprintf(stderr, "%d %s\n", result != 0, fun->name);
*/
    return result;
}

TTterm TTparseTIGER_CONJUNCTION(TTfun fun, TTterm(**parsers) (void))
{
    TTterm result = TTparse_conjunction(fun, parsers);
/*
fprintf(stderr, "%d %s\n", result != 0, fun->name);
*/
    return result;
}

/* The lexical sorts have to be defined from scratch. */
TTterm TTparseTiger_integer( void )
{
    const char * buf;
    size_t len;
    BOOL hasdigit = FALSE;
    int ch;

    buf = TTparse_get_string();

    len = 0;
    ch = buf[len];
    if (ch == '-' || ch == '+')
    {
        len++;
    }
    while (isdigit((int)buf[len]))
    {
        len++;
        hasdigit = TRUE;
    }
    ch = buf[len];
    if (ch == '.' || ch == ',')
    {
        len++;
        if (!isdigit((int)buf[len]))
        {
            len--;
        }
        else
        {
            while (isdigit((int)buf[len]))
            {
                len++;
                hasdigit = TRUE;
            }
        }
    }
    if (hasdigit && !isalnum((int)buf[len]) && buf[len] != '-')
    {
        return TTbuildTiger_integer(scan_word(len));
    }
    return 0;
}

TTterm TTparseTiger_string( void )
{
    const char * buf;
    size_t len, old_len;

    buf = TTparse_get_string();

    len = 0;
    if (buf[len] == '\"')
    {
        len++;
	do {
		old_len = len;
        	len += strcspn(buf+len, "\"");
		if (buf[len] == '\"' && buf[len+1] == '\"') len += 2;
	} while (old_len != len);
        if (buf[len] == '\"')
        {
            len++;
            return TTbuildTiger_string(scan_word(len));
        }
    }
    return 0;
}

TTterm TTparseTiger_id( void )
{
    const char * buf;
    size_t len, i;
    char tmp[MAX_KW_SIZE+1];

    buf = TTparse_get_string();

    len = 0;
    while (isalnum((int)buf[len]) || buf[len] == '_')
    {
        len++;
    }
    if (len > 0)
    {
        if (len < MAX_KW_SIZE)
        {
	    for (i = 0; i < len; i++) {
		tmp[i] = buf[i];
	    }
            tmp[len] = '\0';
            if (TIGERisKeyword(tmp)) return 0;
        }
        return TTbuildTiger_id(scan_word(len));
    }
    return 0;
}


